#!/usr/bin/perl
#
#	CheckLinks.pl		URL extracter and checker script
#
#	Written by John Young, 22 November 2002
#	Copyright (c) 2002 Thistle Software
#
#	Prints a list of URLs from given HTML file(s).
#
# HISTORY
#    Ver	Date		By	Description
#    ---------------------------------------------------------------------------
#    1.0  22-Nov-02	JY	Created
#	2.0   1-Feb-03 JY   Added support for web-site checking.
#	2.1	15-Feb-03 JY	Tidied up the above.


use strict;
use LWP::UserAgent;
use HTTP::Request;
use HTML::LinkExtor;
use URI::URL; # required by HTML::LinkExtor, when invoked with base


my $version = "2.1";
my $scriptName = $0;
#my $scriptName = ($0 =~ m{([^\\/]+[\\/])*(.*?)\.pl\s*$}) ? $2 : $0;


sub usage
{
     print $_[0] if defined ($_[0]);
"\
\
usage:  \
     perl $scriptName.pl [-hprvV] <url>...  > <reportfile>\
\
     where\
          <url>             The URL to check links on.  This should be in the \
                            form http://www.site.com/webpage.html, or the name\
                            of HTML files to check (usually \*.html\").\
          -r                Recurse all sub-directories (only relevant when \
                            a file spec. is specified).\
          -d                Dry - prints out all URLs (but doesn't check them)\
          -b                Report only broken links.\
          -w                Prints an HTML style report.\
          -h                Displays the usage (this message)\
          -v                Verbose.  This output is directed to STDERR.\
          -V                Display version number\
\
     Produces an HTML report on all the links from the given web pages to \
     standard output.  Bad links are highlighted in red.\
\
     If a URL starts with \"http://\", then it this spiders out the entire site\
     (or as much of it as it can reach via links followed recursively from the \
     starting page), checking all HREF and SRC attributes to make sure they \
     work using GET and HEAD requests from LWP::UserAgent. \
\
     If the URL is a file (or file spec.), then the search is not recursive, \
     but only checks the specified files.\n";
}


# Deal with the command line...
use Getopt::Std;
use vars qw ($opt_b $opt_d $opt_h $opt_r $opt_v $opt_V $opt_w);
getopts ('bdhprvVw');
print &usage and exit (0) if $opt_h;

if ( $opt_V )
{
     print "\
     HTML URL extraction utility $version\
     Copyright (c) 2000 Young Associates\
     All rights reserved\n";
     exit (0);
}

my $brokenOnly = $opt_b;
my $dry        = $opt_d;
my $recurse    = $opt_r;
my $verbose    = $opt_v;
my $htmlReport = $opt_w;

print STDERR "Script name is '$scriptName'\n" if $verbose;

my @sources = @ARGV;
print &usage ("No URL or files specified") and exit (1) if !@sources;

# Global variables.
my $email = "jyoung\@gol.com";	# The place to send the report.
my $timeout    = 5;    			# number of seconds to timeout a request
my $MAXPAGES	= 1000;			# Stop after spidering 1000 pages.
my $RQ_DELAY	= 1;    			# number of seconds between requests
my $lastRequest = 0;    			# Time of the last request, for $RQ_DELAY

my $GOODLINK  = 1;
my $BADLINK   = -1;
my $NOTTESTED = 0;


my $startDir = `cd`;
chomp $startDir;

my @urls;
my @files;
foreach my $source (@sources)
{
	if ( $source =~ m{^https?://} )
	{
		push @urls, $source;
	}
	else
	{
		push @files, $source;
	}
}

if ( $verbose )
{
	if ( @files )
	{
		print STDERR "File specs. to check";
		print STDERR " (recursively)" if $recurse;
		print STDERR "...\n";
		foreach my $file (@files)
		{
			print STDERR "   $file\n";
		}
	}
	else
	{
		print STDERR "No file specs given\n";
	}

	if ( @urls )
	{
		print STDERR "URLs to check...\n";
		foreach my $url (@urls)
		{
			print STDERR "   $url\n";
		}
	}
	else
	{
		print STDERR "No URLs given\n";
	}
}




my %Visited;		# Hash of files/URLs checked against a list of links
my %LinkStatus;	# Hash of files/URLs against status...
				# 
				#	0  = Not checked
				#	1  = Good
				#	-1 = Bad


# first, construct the user agent
my $UserAgent = LWP::UserAgent->new;
$UserAgent->agent("$scriptName " . $UserAgent->agent);
$UserAgent->from($email);
$UserAgent->timeout($timeout); # set timeout interval

SpiderURLs (\@urls) if @urls;
CheckFiles (\@files) if @files;

PrintReport ($htmlReport);

exit 0;


sub CheckFiles
{
	my ($specs) = @_;

	my @files = GetFileList ($specs);

	my $base_url;
	
	my $parser = HTML::LinkExtor->new (undef, $base_url);
	foreach my $file ( @files )
	{
		print STDERR "Parsing file '$file'...\n" if $verbose;
		
		if ( !defined $parser->parse_file ($file) )
		{
			print STDERR "   cannot open file\n";
			next
		}

		if ( exists $Visited {$file} )
		{
			warn "Visting $file more than once!\n" if exists $Visited {$file};
		}
		else
		{
			$Visited {$file} = [];
		}

		# $parser->links returns a list of anonymous array with the following
		# elements...
		#
		#   [$tag, $attr => $url, $attr2 => $url2, ...]
		#
		foreach my $linkarray ( $parser->links ) 
		{
			my ($tag, %links) = @{$linkarray};
			if ($tag =~ /^(a|img|frame)$/) 
			{
				while (my($attr, $target) = each %links) 
				{
					#print STDERR "tag=$tag attr=$attr value=$target\n";

					next if $target =~ m|#|;		# Skip relative links

					push @{$Visited {$file}}, $target;
					
					if ( $target =~ /^(?:https?|ftp|mailto):/ )
					{
						# link - check it
						if ( !$dry and $target !~ m{mailto:} )
						{
							my ($success, $type, $actual) = CheckURL ($target);
							$LinkStatus {$target} = $success ? $GOODLINK : $BADLINK;
						}
					}
					else
					{
						# file - good if it exists
						my $exists = -e $target;
						$LinkStatus {$target} = $exists ? $GOODLINK : $BADLINK;
						print STDERR "   checking '$target'..."; 
						print STDERR $exists ? "good" : "bad";
						print STDERR "\n";
					}
				}
			}
		}
	}
	
	if ( $recurse )
	{
		my @dirs = glob ("*");
		foreach my $dir (@dirs)
		{
			if ( -d $dir )
			{
				chdir ($dir);
				my $relDir = GetRelDir ($dir);
				print STDERR "Directory '$relDir'\n" if $verbose;
				CheckFiles ($specs);
				chdir ("..");
			}
		}
	}
}


# SpiderURLs (\@urls);
#
#	Checks each URL in turn by calling SpiderURL for each URL in the list.
#
sub SpiderURLs
{
	my ($urls) = @_;

	for my $url ( @{$urls} )
	{
		SpiderURL ($url);
	}
}

# SpiderURL ($url);
#
#	Checks the given URL for good and bad links.
#

sub SpiderURL
{
	my ($startURL) = @_;
	
	print STDERR "Checking HTTP page '$startURL'\n" if $verbose;
	
	my ($success, $type, $actual) = &CheckURL($startURL);
	
	unless ($success and $type eq 'text/html') 
	{
	    warn "The URL '$startURL' isn't reachable, or isn't HTML.\n";
	    $LinkStatus {$startURL} = $BADLINK;
	    return;
	}
	
	my @urlsQueue;
	my $pageCount;
	
	$Visited {$startURL} = [];
	$LinkStatus {$startURL} = $GOODLINK;
	push @urlsQueue, $startURL;
	my $startBase = $startURL;
	$startBase    =~ s{/[^/]*$}{/}; # trim everything after last '/'
	
	while (@urlsQueue) 
	{
	    ++$pageCount;
	    if ($pageCount > $MAXPAGES) 
	    {
	        warn "stopped checking after reaching $MAXPAGES pages.\n";
	        --$pageCount; # kludge so the count is correct in report
	        last;
	    }
	    my $page = shift @urlsQueue;
	    my @newPages = &ProcessPage ($page, $startBase);
	    push @urlsQueue, @newPages;
	}
}


# PrintReport ($html);
#
# 	Produces a report of the good and bad links found.  If $html is 1 then an
#	HTML report is produces, otherwise a plain text report is produced.
sub PrintReport
{	
	my ($htmlReport) = @_;
	
	if ( $htmlReport ) 
	{
		PrintHTMLReport();
	}
	else
	{
		PrintTextReport();
	}
}


# PrintHTMLReport();
#
# 	Print an HTML version of the report.
#
sub PrintHTMLReport
{
	my $startURL = "PLACEHOLDER";
	my $time = localtime;
    
	    print <<EndOfHeader;
<html>
<head>
<title>Link check report</title>
</head>
<body>

<H1>Link check report</H1>

<P><STRONG>Report created at $time</STRONG>

<P>
EndOfHeader
	
	foreach my $file (sort keys %Visited) 
	{
		# Get rid of dups and non-broken links (if asked)
		my @links;
		my $prev;
		for my $link (sort @{$Visited{$file}})
		{
			if ( $prev ne $link )
			{
				push @links, $link if !$brokenOnly or $LinkStatus {$link} eq $BADLINK;
				$prev = $link;
			}
		}

		my $count = @links;
		next if $count == 0 and $brokenOnly;
		
		print "<p><strong><a href=\"$file\">$file</A> ($count links)</strong><br>\n";
		print "<table width=\"90%\" align=\"center\">\n";
		for my $link (@links)
		{
			my $state  = $LinkStatus {$link};
	
			print "    <tr><td width=\"70%\"><a href=\"$link\">$link</a>";
			
			if ( $state == 0 )
			{
				print "        <td><font color=\"gray\">not tested</font>";
			}
			elsif ( $state == -1 )
			{
				print "        <td><font color=\"red\"><b>BROKEN!</b></font>";
			}
			else
			{
				print "        <td>good";
			}

			print "\n";
		}
		print "</table>\n\n\n";
	}
	print "</BODY></HTML>\n";
}



# just print a plain-text version of the report
#
#
sub PrintTextReport
{	

#	my $startURL = "PLACEHOLDER";
#	my $time = localtime;
#
#	print "$startURL $0 report\n";
#	print "Report created at $time\n\n";
#	
#	foreach my $file (sort keys %BadLinks) 
#	{
#		print "$file:\n";
#		foreach my $target (sort @{ $BadLinks{$file} }) 
#		{
#			print " $target\n";
#		}
#		print "\n";
#	} 
}



# @links = ProcessPage ($page, $startBase)
#
#	Invoked with a single argument of a page under the $startBase that needs 
#	to be processed. That page will be (1) retrieved via GET, (2) parsed for 
#	any links it contains, and (3) have those links checked for validity 
#	themselves (with any that fail being written to %BadLinks, and any that
# 	point to valid, unchecked HTML files under $startBase being added to 
#	@urlsQueue). The subroutine has no return value.
#
sub ProcessPage 
{
	my ($page, $startBase) = @_;
	return unless defined $page;
	
	warn "processing '$page' for links\n" if $verbose;

	sleep 1 while (time - $lastRequest) < $RQ_DELAY;
	$lastRequest = time;
	my $response = $UserAgent->request(HTTP::Request->new('GET', $page));
	my $contType = $response->header('Content-Type');
	unless ($response->is_success and 
		   $contType =~ m{^text/html} )
	{
		# strange, since it passed these tests
		# via HEAD request in order to get into @urlsQueue
		#undef $Visited {$page};
		print STDERR "WARNING: Strange, since '$page' passed these tests via HEAD previously\n";
		my $success = $response->is_success;
		my $contType = $response->header('Content-Type');
		print STDERR "reponse->success=$success Content-Type='$contType\n";
		return;
	}
	
	my $base = $response->base;
	
	my $escapedStartBase = quotemeta $startBase;
	unless ($base =~ /$escapedStartBase/o) 
	{
		# looks like we were redirected away from $startBase
		return;
	}


	my @urlsQueue;

	my $parser = HTML::LinkExtor->new(undef, $base);
	$parser->parse($response->content);
	
	# $parser->links returns a list of anonymous array with the following
	# elements...
	#
	#   [$tag, $attr => $url, $attr2 => $url2, ...]
	#
	foreach my $linkarray ( $parser->links ) 
	{
		my ($tag, %links) = @{$linkarray};
		if ($tag =~ /^(a|img|frame)$/) 
		{
	
		TARGET: 
			while (my($attr, $target) = each %links) 
			{
				if ($attr =~ /^(href|src|lowsrc)$/) 
				{
					# these $target entries are the ones we're
					# interested in.
					next TARGET unless $target =~ /^(?:https?|ftp|mailto):/;
					$target =~ s/#.*//; # lose trailing #targets

					push @{$Visited {$page}}, $target;

					if ( !exists $LinkStatus {$target} ) # have already seen this before?
					{
						# haven't seen this one before
						if ( !$dry and $target !~ m{^mailto:} )
						{
							my ($success, $type, $actual) = &CheckURL ($target);
							$LinkStatus {$target} = $success ? $GOODLINK : $BADLINK;
						
							if ( $success and 
								defined $type and 
								$type eq 'text/html' and 
								defined $actual and 
								$actual =~ /$escapedStartBase/o ) 
							{
								push @urlsQueue, $target;
							}
						}
					}
					elsif ( $verbose )
					{
						print STDERR "   already checked '$target' ";
						if ( $LinkStatus {$target} == $GOODLINK )
						{
							print STDERR "(good)\n";
						}
						elsif ( $LinkStatus {$target} == $BADLINK )
						{
							print STDERR "(bad)\n";
						}
						else
						{
							print STDERR "(not tested)\n";
						}
					}
				}
			}
		}
	}
	return @urlsQueue;
}



# CheckURL ($url)
#
#	Check that this URL is valid, using the HEAD (and GET, if HEAD fails) 
#	methods. Returns a 3-element array: ($success, $type, $actual)...
#
#		$success = 1 if the URL is valid
#		$type is the Content-type (usually 'text/html')
#		$actual is the base URI
#
sub CheckURL
{
	my $url = shift;
	
	print STDERR "   checking '$url'..." if $verbose; 
	
	unless (defined $url) 
	{
		return ('', '', '');
	}
	
	# Wait until it's time to send another request.
	sleep 1 while (time - $lastRequest) < $RQ_DELAY;
	
	$lastRequest = time;
	my $response = $UserAgent->request(HTTP::Request->new('HEAD', $url));
	my $success = $response->is_success;
	unless ($success) 
	{
		# try a GET request; some hosts don't like HEAD
		sleep 1 while (time - $lastRequest) < $RQ_DELAY;
		$lastRequest = time;
		$response = $UserAgent->request(HTTP::Request->new('GET', $url));
		$success = $response->is_success;
	}

	if ($verbose) 
	{
		if ($success) 
		{
			warn "good.\n";
		} 
		else 
		{
			warn "bad.\n";
		}
	}

	my $type = $response->header('Content-Type');
	my $actual;
	if ($success) 
	{
		$actual = $response->base; 	# redirected?
	}
	return ($success, $type, $actual);
}






sub GetFileList
{
	my $filespecs = @_[0];
	my @list;
	foreach my $spec (@{$filespecs})
	{
		$spec =~ s%\\%/%g;			# Replace backslashes with forward slashes
		$spec =~ s% %\\ %g;
		push @list, glob ($spec);
	}
	return @list;
}



# $relDir = GetRelDir ($dir);
#
#	Returns the current directory relative to $startDir (which is set at the 
#	start of the script).  If there is something wrong, then the default ($dir) 
#	is returned, or the full path if $dir is not given.
#
sub GetRelDir
{
	my ($dir) = @_;
	my $fullDir = `cd`;
	chomp $fullDir;

	#print "Full dir = $fullDir (start='$startDir')\n";

	my $relDir;
	if ( $fullDir eq $startDir )
	{
		$relDir = "";
	}
	elsif ( $fullDir =~ /^\Q$startDir\E\\(.*)/ )
	{
		$relDir = $1;
	}
	elsif ( defined $dir )
	{
		$relDir = $dir;
	}
	else 
	{
		$relDir = $fullDir;
	}

	#print "   Returning '$relDir'\n";
	return $relDir;
}